/*
 *	WAP irc daemon
 *	(c) 2001 Alan Cox <alan@lxorguk.ukuu.org.uk>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Due to patent restrictions on WAP in the USA this software is not licensed
 *  for use within the United States of America, its dependancies, embassies
 *  or other jurisdictions
 */
 
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <ctype.h>
#include <signal.h>

#include <netinet/in.h>
#include <sys/socket.h>

#include <arpa/inet.h>
#include <netdb.h>
#include <errno.h>
#include <sys/time.h>

#include <sys/un.h>

#include <sys/stat.h>

#include "config.h"
#include "cgi.h"
#include "web_tty.h"

static char data[8192];
static int dptr=0;
static int dead=1;
static int ircfd;

static int host_connect_attempt(struct in_addr ia, int port)
{
	int s=socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
	
	struct sockaddr_in sin;
	
	fd_set wfd;
	struct timeval tv;
	
	if(s==-1)
	{
		perror("socket");
		return -1;
	}
	
	if(fcntl(s, F_SETFL, FNDELAY)==-1)
	{
		perror("nonblocking");
		close(s);
		return -1;
	}

	sin.sin_family = AF_INET;	
	sin.sin_addr   = ia;
	sin.sin_port   = htons(port);
	
	if(connect(s, (struct sockaddr *)&sin, sizeof(sin))==-1 && errno != EINPROGRESS)
	{
		perror("connect");
		close(s);
		return -1;
	}	
	
	tv.tv_sec = 60;		/* We use 60 second timeouts for now */
	tv.tv_usec = 0;
	
	FD_ZERO(&wfd);
	FD_SET(s, &wfd);
	
	switch(select(s+1, NULL, &wfd, NULL, &tv))
	{
		case 0:
			/* Time out */
			close(s);
			return -1;
		case -1:
			/* Ermm.. ?? */
			perror("select");
			close(s);
			return -1;
	}
	
	return s;
}

int host_connect(const char *host, int port)
{
	struct hostent *h;
	int i;
	int s;
	
	h=gethostbyname(host);
	if(h==NULL)
	{
		fprintf(stderr,"unable to resolve '%s'.\n", host);
		return -1;
	}
	
	
	for(i=0; h->h_addr_list[i]; i++)
	{
		struct in_addr ia;
		memcpy(&ia, h->h_addr_list[i],4);
		s=host_connect_attempt(ia, port);
		if(s != -1)
			return s;
	}
	fprintf(stderr, "unable to connect to '%s'.\n", host);
	return -1;
}
	
static void wait_message(int cmdfd)
{
	fd_set in;
	int n=ircfd;
	struct timeval tv;
	
	FD_ZERO(&in);
	if(!dead)
	{
		FD_SET(ircfd, &in);
		FD_SET(1, &in);
	}
	FD_SET(cmdfd, &in);

	if(cmdfd>ircfd)
		n=cmdfd;
		
	tv.tv_sec = 180;
	tv.tv_usec = 0;
	
	if(select(n+1, &in, NULL, NULL, dead?NULL:&tv)==0 && !dead)
	{
		close(ircfd);
		ircfd = -1;
		dead=1;
	}
	
	if(FD_ISSET(cmdfd, &in))
	{
		int tfd=accept(cmdfd, NULL, 0);
		int l;
		char buf[256];
		if(tfd==-1)
			return;
			
		if(dead)
		{
			print_out("\n\nConnecting as ");
			print_out("Alan7110");
			print_out("... \n");

			ircfd=host_connect(IRC_SERVER_NAME, 6667);
			if(ircfd!=-1)
				dead=0;
			irc_setup(ircfd);
		}
		l=read(tfd, buf, 255);
		buf[l]=0;
		if(*buf=='>')
			parseinput(buf+1);
		if(dptr<1200)
		{
			write(tfd, data, dptr);
			dptr=0;
		}
		else
		{
			write(tfd, data, 1200);
			dptr-=1200;
			memmove(data, data+1200, dptr);
			write(tfd, "...",3);
		}
		close(tfd);
	}
}

/*
 *	Lost link 
 */

void print_out(char *p)
{
	int l = strlen(p);
	if(l+dptr>8191)
		exit(1);
	memcpy(data+dptr, p, l);
	dptr+=l;
	printf("%s", p);
}

void print_bit(char c)
{
	if(dptr==8191)
		exit(1);
	data[dptr++]=c;
	printf("%c",c);
}
	

static void read_message()
{
	int l=server_input();	
	if(l==0)
	{
		print_out("\n**Connection Closed!\n\n");
		dead=1;
	}
}


static int make_unix_socket(void)
{
	struct sockaddr_un sun;
	int fd;
	char name[256];

	snprintf(name, 	256, WAPIRC_SOCKPATH);
	unlink(name);
	
	sun.sun_family=AF_UNIX;
	strcpy(sun.sun_path, name);
	
	fd=socket(PF_UNIX, SOCK_STREAM, 0);
	if(fd==-1)
	{
		perror("socket");
		exit(1);
	}
	
	if(umask(077)==-1)
	{
		perror("umask");
		exit(1);
	}

	if(bind(fd, (struct sockaddr *)&sun, sizeof(sun))!=0)
	{
		perror("bind");
		exit(1);
	}

	if(chown(WAPIRC_SOCKPATH, WEB_USERID, -1))
	{
		perror("chown");
		exit(1);
	}

	if(chmod(WAPIRC_SOCKPATH, 0777))
	{
		perror("chmod");
		exit(1);
	}
	
	listen(fd,2);
	
	return fd;
}

		
int main(int argc, char *argv[])
{
	int cmdfd;
	signal(SIGPIPE, SIG_IGN);
	
	cgi_process();
		
	cmdfd=make_unix_socket();

	while(1)
	{
		wait_message(cmdfd);
		read_message();
	}
}
	
